package com.android.demo.serialportservice;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.concurrent.atomic.AtomicInteger;

import android.app.Activity;
import android.app.AlertDialog;
import android.hardware.IOManager;
import android.hardware.SerialManager;
import android.hardware.SerialPort;
import android.os.Bundle;
import android.os.Handler;
import android.os.Message;
import android.util.Log;
import android.view.View;
import android.view.View.OnClickListener;
import android.widget.AdapterView;
import android.widget.AdapterView.OnItemSelectedListener;
import android.widget.ArrayAdapter;
import android.widget.Button;
import android.widget.CheckBox;
import android.widget.EditText;
import android.widget.ScrollView;
import android.widget.Spinner;
import android.widget.TextView;
import android.widget.Toast;

public class SerialPortActivity extends Activity implements OnClickListener {

	private static final String TAG = "SerialPortActivityTag";

	private static final int DEFAULT_BUADRATE = 9600;
	private static final String DEFAULT_DEVICE = "/dev/ttyS1";
	private static final int MSG_SHOW=100;

	private String mRTSs[] = { "CLR", "SET" };
	private String mDeviceNamePaths[];
	private String mBaudrates[];
	private String mFlowCtrls[];

	private byte mBaudrateIndex, mDeviceIndex, mSetRTSIndex, mFlowCtrlIndex;

	Runnable mSendDataRunnable = new Runnable() {
		public void run() {
			sendData();
		}
	};

	private void sendData() {
		Log.d(TAG, "sendData() called.");
		if(mContinueTest) {
			synchronized (mSyncSerial) {
				if (mSerialPort == null) {
					mContinueTest = false;
					return;
				}
			}

			//Send 256 bytes
			try {
				byte[] cmds = new byte[256];
				for (int i = 0; i < 256; i++) {
					cmds[i] = (byte) (i & 0xFF);
				}

				ByteBuffer buffer = ByteBuffer.wrap(cmds);
				mSerialPort.write(buffer, cmds.length);
			}
			catch (IOException e) {
				Log.e(TAG, "Send data failure: " + e);
			}
			//@}
			mHandler.postDelayed(mSendDataRunnable, 1000);
		}
	}

	// private TextView mTxMode;
	private ScrollView scroller;
	private TextView mReception;
	private EditText mCmdArea, mCmdArea2;
	private Button mBtnClear, mBtnSend, mBtnOpenClose, mBtnChangeMode, mBtnSend2, mBtnTest;

	private CheckBox mCheckHexMode;
	private Spinner mSpDevice, mSpFlowCtrl, mSpBaudRate, mSpMode;

	private boolean mContinueTest;
	private SerialManager mSerialMgr;
	private SerialPort mSerialPort;
	private ReadThread mReadThread;
	private IOManager mIOManager;
	private Object mSyncSerial = new Object();
	private Handler mHandler = new Handler() {
		public void handleMessage(Message msg) {

			switch (msg.what) {
				case MSG_SHOW:
					String str = (String) msg.obj;
					if (mReception != null) {
						if(mReception.length() > 2000)
							mReception.setText("");//clear

						mReception.append(str);
						scrollDown();
					}
					break;
			}
		}
	};

	private class ReadThread extends Thread {

		@Override
		public void run() {
			Log.d(TAG, "Entering ReadThread (" + this.getId() + ")");

			ByteBuffer mBuffer = ByteBuffer.allocate(512);
			int size = 0;
			int getData = 0;

			super.run();
			while (!isInterrupted()) {

				try {
					synchronized (mSyncSerial) {
						if (mSerialPort == null)
							break;
						else {
							size = mSerialPort.read(mBuffer);
						}
						
						if (size == 0)
							mSyncSerial.wait(50);
					}

					if (size > 0) {
						// Log.d(TAG, "onDataReceived");
						onReceived(mBuffer.array(), size);
					}
				} catch (IOException e) {
					Log.d(TAG, "IOException");
					mContinueTest = false;
					break;
				} catch (Exception e) {
					// e.printStackTrace();
					Log.d(TAG, "ReadThread exception");
					mContinueTest = false;
					break;
				}
			}// while

			Log.d(TAG, "Leaving ReadThread (" + this.getId() + ")");
		}
	}

	/** Called when the activity is first created. */
	@Override
	public void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);
		setContentView(R.layout.main);

		mSerialMgr = (SerialManager) getSystemService("serial");
		mIOManager = (IOManager) getSystemService("gpio");

		String deviceSerialNumber = mIOManager.getDeviceSerialNumber();
		
		try {
			mDeviceNamePaths = mSerialMgr.getSerialPorts();
			if (mDeviceNamePaths == null)
				mDeviceNamePaths = new String[]{DEFAULT_DEVICE};
		} catch (SecurityException e) {

			Log.e(TAG, "Security Exception");
			Log.e(TAG, e.getMessage());

		}catch (NullPointerException e) {
			toastMsg("Serial service not found !");
			Log.e(TAG, "Serial service not found");
		}
		catch (Exception e) {
			Log.e(TAG, e.getMessage());
		}
		if (mDeviceNamePaths == null)
			mDeviceNamePaths = new String[]{"NOT Supported"};

		mBaudrates = getResources().getStringArray(R.array.baudrates_name);

		mFlowCtrls = getResources().getStringArray(R.array.flowctrl);

		mSpDevice = (Spinner) findViewById(R.id.deviceName);
		mSpFlowCtrl = (Spinner) findViewById(R.id.flowctrolSelection);
		mSpBaudRate = (Spinner) findViewById(R.id.baudrateSelection);
		mSpMode = (Spinner) findViewById(R.id.rtsSelection);

		// mTxMode = (TextView) findViewById(R.id.mode);

		scroller = (ScrollView) findViewById(R.id.scroller);

		// Edit area
		mCmdArea = (EditText) findViewById(R.id.EditEmission);
		mCmdArea2 = (EditText) findViewById(R.id.EditEmission2);
		mReception = (TextView) findViewById(R.id.messageView);

		// Buttons
		mBtnOpenClose = (Button) findViewById(R.id.btnOpenClose);
		mBtnOpenClose.setOnClickListener(this);

		mBtnChangeMode = (Button) findViewById(R.id.btnChangeMode);
		mBtnChangeMode.setOnClickListener(this);

		mBtnSend = (Button) findViewById(R.id.btnSend);
		mBtnSend.setOnClickListener(this);

		mBtnSend2 = (Button) findViewById(R.id.btnSend2);
		mBtnSend2.setOnClickListener(this);

		//JR: Add Send Test
		mBtnTest = (Button) findViewById(R.id.btnTest);
		mBtnTest.setOnClickListener(this);
		//@}

		mBtnClear = (Button) findViewById(R.id.btnClear);
		mBtnClear.setOnClickListener(this);

		// Check box
		mCheckHexMode = (CheckBox) findViewById(R.id.hexMode);

		mCheckHexMode.setChecked(true);

		initDeviceList();
		initFlowCtrlList();
		initBaudrateList();
		initRTSCtrl();

//		mCmdArea.setText("04A30400FF55");
		mCmdArea.setText("30313233340D");
		mCmdArea2.setText("7777");

		findViewById(R.id.btn5vOn).setOnClickListener(v -> gpioWrite(70, 1));
		findViewById(R.id.btn5vOff).setOnClickListener(v -> gpioWrite(70, 0));

		findViewById(R.id.btn3vOn).setOnClickListener(v -> gpioWrite(135, 1));
		findViewById(R.id.btn3vOff).setOnClickListener(v -> gpioWrite(135, 0));

		findViewById(R.id.btnRstOn).setOnClickListener(v -> gpioWrite(82, 1));
		findViewById(R.id.btnRstOff).setOnClickListener(v -> gpioWrite(82, 0));
	}

	private void gpioWrite(int gpio, int value) {
		boolean success = mIOManager.gpioWrite(gpio, value);
		Toast.makeText(this, success ? "Success" : "Failed", Toast.LENGTH_SHORT).show();
	}

	void initFlowCtrlList() {
		ArrayAdapter<String> adapter = new ArrayAdapter<String>(this,
				android.R.layout.simple_spinner_item, mFlowCtrls);

		adapter.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);

		mSpFlowCtrl.setAdapter(adapter);
		mSpFlowCtrl.setOnItemSelectedListener(new OnItemSelectedListener() {
			public void onItemSelected(AdapterView<?> parent, View view,
					int position, long id) {

				mFlowCtrlIndex = (byte) (id);// id index based from 0

				Log.d(TAG, "FlowCtrl =" + mFlowCtrls[mFlowCtrlIndex]);

				setFlowCtrl(mFlowCtrlIndex + 1);
			}

			public void onNothingSelected(AdapterView<?> parent) {
			}
		});

		mSpFlowCtrl.setSelection(mFlowCtrlIndex);
	}

	void initRTSCtrl() {
		ArrayAdapter<String> adapter = new ArrayAdapter<String>(this,
				android.R.layout.simple_spinner_item, mRTSs);

		adapter.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);

		mSpMode.setAdapter(adapter);
		mSpMode.setOnItemSelectedListener(new OnItemSelectedListener() {
			public void onItemSelected(AdapterView<?> parent, View view,
									   int position, long id) {

				mSetRTSIndex = (byte) (id);// id index based from 0

				Log.d(TAG, "RTS Enable= " + mSetRTSIndex);

				setRTS(mSetRTSIndex);
			}

			public void onNothingSelected(AdapterView<?> parent) {
			}
		});

		mSetRTSIndex = 0;
		mSpMode.setSelection(mSetRTSIndex);
	}

	void initDeviceList() {
		ArrayAdapter<String> adapter = new ArrayAdapter<String>(this,
				android.R.layout.simple_spinner_item, mDeviceNamePaths);

		adapter.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);

		mSpDevice.setAdapter(adapter);
		mSpDevice.setOnItemSelectedListener(new OnItemSelectedListener() {
			public void onItemSelected(AdapterView<?> parent, View view,
									   int position, long id) {

				mDeviceIndex = (byte) (id);// id index based from 0

				Log.d(TAG, "Device =" + mDeviceNamePaths[mDeviceIndex]);

			}

			public void onNothingSelected(AdapterView<?> parent) {
			}
		});

		mDeviceIndex = 0; // init

		// Look for DEFAULT_DEVICE
		for (int i = 0; i < mDeviceNamePaths.length; i++) {
			if (mDeviceNamePaths[i].endsWith(DEFAULT_DEVICE)) {
				mDeviceIndex = (byte) i;
				break;
			}
		}
		mSpDevice.setSelection(mDeviceIndex);

	}

	void initBaudrateList() {
		ArrayAdapter<String> adapter = new ArrayAdapter<String>(this,
				android.R.layout.simple_spinner_item, mBaudrates);

		adapter.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);

		mSpBaudRate.setAdapter(adapter);
		mSpBaudRate.setOnItemSelectedListener(new OnItemSelectedListener() {
			public void onItemSelected(AdapterView<?> parent, View view,
					int position, long id) {

				mBaudrateIndex = (byte) (id);// id index based from 0

				Log.d(TAG, "Baudrate =" + mBaudrates[mBaudrateIndex]);

				// change baudrate
				setBaudrate(Integer.valueOf(mBaudrates[mBaudrateIndex]));
			}

			public void onNothingSelected(AdapterView<?> parent) {
			}
		});

		// Look for baudrate
		for (int i = 0; i < mBaudrates.length; i++) {
			if (Integer.valueOf(mBaudrates[i]) == DEFAULT_BUADRATE) {
				mBaudrateIndex = (byte) i;
				break;
			}
		}
		mSpBaudRate.setSelection(mBaudrateIndex);
	}

	/**
	 * Called after onRestoreInstanceState(Bundle), onRestart(), or onPause(),
	 * for your activity to start interacting with the user. This is a good
	 * place to begin animations, open exclusive-access devices (such as the
	 * camera), etc.
	 * 
	 * Derived classes must call through to the super class's implementation of
	 * this method. If they do not, an exception will be thrown.
	 */
	@Override
	protected void onResume() {
		Log.i(TAG, "onResume()");

		super.onResume();
	}

	public void setBaudrate(int speed) {

		synchronized (mSyncSerial) {
			if (mSerialPort == null)
				return;
		}

		try {
			mSerialPort.setBaudrate(speed);
		} catch (NumberFormatException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (Exception e) {
			e.printStackTrace();
		}

	}

	public void setRTS(int set) {

		synchronized (mSyncSerial) {
			if (mSerialPort == null)
				return;
		}

		try {
			if (mSerialPort != null)
				mSerialPort.setRTS(set);
		} catch (IOException e) {
			e.printStackTrace();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}


	public void setFlowCtrl(int flowType) {
		synchronized (mSyncSerial) {
			if (mSerialPort == null)
				return;
		}

		if (mSerialPort != null) {
			// set prefer flow ctrl
			try {
				mSerialPort.setFlowCtrl(flowType);
			} catch (IOException e) {
				e.printStackTrace();
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

	}

	public boolean openSerialPort() {

		try {
			synchronized (mSyncSerial) {
				mSerialPort = mSerialMgr.openSerialPort(
						mDeviceNamePaths[mDeviceIndex],
						Integer.valueOf(mBaudrates[mBaudrateIndex]));
			}

		} catch (Exception e) {
			Log.d(TAG, e.getMessage());
			return false;
		}

		if (mSerialPort != null) {

			if (mReadThread == null || !mReadThread.isAlive()) {
				mReadThread = new ReadThread();
				mReadThread.start();
			}

			setRTS(mSetRTSIndex); // platfrom/driver dependent
			setFlowCtrl(mFlowCtrlIndex + 1);// platfrom/driver dependent

			mBtnOpenClose.setText(R.string.close);

			mSpDevice.setEnabled(false);

			return true;

		} else {
			return false;
		}

	}

	public boolean closeSerialPort() {
		boolean ok = false;

		if (mReadThread != null)
			mReadThread.interrupt();

		try {
			synchronized (mSyncSerial) {
				if (mSerialPort != null) {
					mSerialPort.close();
					ok = true;
				}
			}

			// Waiting for the reading thread is terminated
			if (mReadThread != null && mReadThread.isAlive()) {
				try {
					mReadThread.join();
				} catch (InterruptedException e) {
					// e.printStackTrace();
				}
			}
		} catch (IOException e) {
			Log.d(TAG, e.getMessage());
		}
		catch (Exception e) {
			Log.d(TAG, e.getMessage());
		}
		finally {
			synchronized (mSyncSerial) {
				mSerialPort = null;
			}
		}

		return ok;
	}

	@Override
	public void onPause() {
		Log.i(TAG, "onPause()");
		super.onPause();

		closeSerialPort();
	}

	@Override
	protected void onDestroy() {
		Log.i(TAG, "onDestroy()");

		// MUST CALL !!!
		super.onDestroy();
	}

	/**
	 * Called when you are no longer visible to the user. You will next receive
	 * either {@link #onStart}, {@link #onDestroy}, or nothing, depending on
	 * later user activity.
	 */
	@Override
	protected void onStop() {
		Log.i(TAG, "onStop()");
		super.onStop();

	}

	public void onClick(View v) {

		try {
			switch (v.getId()) {
			case R.id.btnClear:
				mReception.setText("");// Clear
				break;

			case R.id.btnOpenClose:

				if (!mSpDevice.isEnabled()) {
					closeSerialPort();

					mBtnOpenClose.setText(R.string.open);

					mSpDevice.setEnabled(true);

				} else {
					// 2) open port

					if (openSerialPort()) {
						mBtnOpenClose.setText(R.string.close);

						mSpDevice.setEnabled(false);
					} else {
						toastMsg(String.format("Couldn't open %s",
								mDeviceNamePaths[mDeviceIndex]));
					}

				}
				break;

			case R.id.btnChangeMode:
				final ArrayList<SerialManager.SERIAL_MODE> modes = mSerialMgr.getSupportSerialModes(mDeviceNamePaths[mDeviceIndex]);
				SerialManager.SERIAL_MODE current = mSerialMgr.getSerialMode(mDeviceNamePaths[mDeviceIndex]);
				CharSequence[] entries = new CharSequence[modes.size()];
				AtomicInteger selectedIndex = new AtomicInteger();

				for(int i = 0; i < modes.size(); i++) {
					if (modes.get(i) == current) {
						selectedIndex.set(i);
					}

					entries[i] = modes.get(i).name();
				}

				new AlertDialog.Builder(this)
						.setTitle("Select mode for " + mDeviceNamePaths[mDeviceIndex])
						.setSingleChoiceItems(entries, selectedIndex.get(), (dialog, which) -> selectedIndex.set(which))
						.setPositiveButton(R.string.ok, (dialog, which) -> {
							SerialManager.SERIAL_MODE mode = modes.get(selectedIndex.get());
							boolean success = mSerialMgr.setSerialMode(mDeviceNamePaths[mDeviceIndex], mode);
							toastMsg("Set " + mDeviceNamePaths[mDeviceIndex] + " mode to " + mode.name() + " " + (success ? "successful" : "failed"));
						})
						.setNegativeButton(R.string.cancel, null)
						.show();
				break;

			case R.id.btnTest:
				sendTest();
				break;
			//@}

			case R.id.btnSend:
			case R.id.btnSend2:

				if (mSerialPort != null) {
					String cmd;

					if (v.getId() == R.id.btnSend)
						cmd = mCmdArea.getText().toString();
					else
						cmd = mCmdArea2.getText().toString();

					if (cmd.length() != 0) {
						byte cmds[];

						if (!mCheckHexMode.isChecked()) {
							cmds = cmd.getBytes();

							ByteBuffer buffer = ByteBuffer.wrap(cmds);
							mSerialPort.write(buffer, cmds.length);

							buffer.clear();
							buffer.put((byte) 0x0D);
							mSerialPort.write(buffer, 1); // return char
						} else {
							cmd = cmd.trim();
							String cmdArray[] = cmd.split(" ");

							cmd = "";
							for (int i = 0; i < cmdArray.length; i++) {
								cmd += cmdArray[i].toUpperCase();
							}

							cmds = stringToHex(cmd);

							ByteBuffer buffer = ByteBuffer.wrap(cmds);

							mSerialPort.write(buffer, cmds.length);
						}

						mReception.append("SEND > ");

						if (!mCheckHexMode.isChecked()) {
							mReception.append(cmd);
							mReception.append("<CR>");
						} else {
							mReception
									.append(hexToString(cmds, cmds.length, 0));
						}

						mReception.append("\n");
						scrollDown();
					}
				} else {
					toastMsg(R.string.portnotopened);
				}

				break;
			}
		} catch (Exception e) {
			Log.d(TAG, "onClick() exception!");
		}
	}

	private void sendTest() {
		if (mSerialPort != null) {
			if(!mContinueTest) {
				//Start TEST
				mContinueTest = true;
				mBtnOpenClose.setEnabled(false);
				mBtnTest.setText(R.string.stop);

				sendData();
			}
			else {
				//STOP TEST
				mContinueTest = false;
				mBtnOpenClose.setEnabled(true);
				mBtnTest.setText(R.string.test);
			}
		}
		else {
			toastMsg(R.string.portnotopened);
		}
	}

	private String hexToString(byte[] buf, int size, int start) {
		StringBuffer strBuf = new StringBuffer(size * 2);
		int i;

		for (i = start; i < start + size; i++) {
			if (((int) buf[i] & 0xff) < 0x10) {
				strBuf.append("0");
			}
			strBuf.append(Integer.toString((int) buf[i] & 0xff, 16)
					.toUpperCase());

		}
		return strBuf.toString();
	}

	private byte[] stringToHex(String str) {
		byte[] bb = str.getBytes();
		byte[] strHex = new byte[2];

		byte[] target;

		if (str.length() % 2 == 0)
			target = new byte[str.length() / 2];
		else
			target = new byte[str.length() / 2 + 1];

		int j = 0;

		for (int i = 0; i < str.length(); i += 2) {

			if (i + 1 < bb.length) {
				strHex[0] = bb[i];
				strHex[1] = bb[i + 1];
			} else if (i < bb.length) {
				strHex[0] = '0';
				strHex[1] = bb[i];
			}

			target[j++] = (byte) Integer.parseInt(new String(strHex), 16);
		}

		return target;
	}

	public void onReceived(byte[] buffer, int size) {
		String strMsg = hexToString(buffer, size, 0);

		Log.d(TAG,
				String.format("RECV: %s (%d bytes)",
						strMsg, size));

		if (!mCheckHexMode.isChecked()) {
			strMsg = new String(buffer, 0, size);
		}

		mHandler.obtainMessage(MSG_SHOW, (Object) strMsg).sendToTarget();
	}

	protected void toastMsg(String msg) {
		Toast.makeText(this, msg, Toast.LENGTH_SHORT).show();
	}

	protected void toastMsg(int resId) {
		String msg = getResources().getString(resId);
		Toast.makeText(this, msg, Toast.LENGTH_SHORT).show();
	}

	private void scrollDown() {
		scroller.smoothScrollTo(0, mReception.getBottom());
	}
}
